/*                                                                -*- mode: c; c-basic-offset: 8;  -*-
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 ============================================================================
 Name        : service_discovery.c
 Version     : 1.0.0
 Copyright   : @CERN, IT-DM/GT; see LICENSE.txt file
 Description : in C, Ansi-style
 : See the ServiceDiscovery.h file
 : for a full description of the functions
 Remarks	 :
 ============================================================================
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef __GNUC__
#define __attribute__(x)
#endif

#include <stdio.h>
#include "ServiceDiscoveryIfce.h"

/**********************************************************************
 * Constants
 */

/* LDAP attributes */
#define ATTR_OC			"objectClass"
#define ATTR_DN			"dn"
#define ATTR_NAME		"GlueServiceUniqueID"
#define ATTR_TYPE		"GlueServiceType"
#define ATTR_VERSION		"GlueServiceVersion"
#define ATTR_LINK		"GlueForeignKey"
#define ATTR_SITE		"GlueSiteUniqueID"
#define ATTR_ENDPOINT		"GlueServiceEndpoint"
#define ATTR_VERSION		"GlueServiceVersion"
#define ATTR_STATUS     	"GlueServiceStatus"
#define ATTR_WSDL		"GlueServiceWSDL"
#define ATTR_ADMIN
#define ATTR_VO			"GlueServiceOwner"
#define ATTR_DATAKEY		"GlueServiceDataKey"
#define ATTR_DATAVALUE		"GlueServiceDataValue"

// in gfal used attributes
#define ATTR_CE_AP_ATNM		"GlueCESEBindCEAccesspoint"
#define ATTR_PROTO_PORT 	"GlueSEAccessProtocolPort"
#define	ATTR_PROTO_TYPE		"GlueSEAccessProtocolType"
#define ATTR_SA_PATH_ATNM 	"GlueSAPath"
#define ATTR_SA_ROOT_ATNM	"GlueSARoot"
#define ATTR_PATH_ATNM 		"GlueVOInfoPath"
#define ATTR_SA_KEY_ATNM 	"GlueChunkKey"
#define ATTR_SE_TYPE_ATPT	"GlueSEPort"
#define ATTR_SE_TYPE_ATST	"GlueSEStatus"

// nonstandard glue stuff to fix LCG inconsistencies
#define ATTR_VERSION_MAJOR	"GlueServiceMajorVersion"
#define ATTR_VERSION_MINOR	"GlueServiceMinorVersion"
#define ATTR_VERSION_PATCH	"GlueServicePatchVersion"
#define ATTR_URI		"GlueServiceURI"
#define ATTR_NM			"GlueServiceName"
#define ATTR_ACCESSRULE     	"GlueServiceAccessControlRule"
#define ATTR_HOSTINGORG     	"GlueServiceHostingOrganization"
#define ATTR_SA_ROOT        	"GlueSARoot"
#define ATTR_SA_ACBR        	"GlueSAAccessControlBaseRule"
#define ATTR_SE	        	"GlueSEUniqueID"
#define ATTR_CHUNK		"GlueChunkKey"

/* LDAP object classes */
#define CLASS_SERVICE		"GlueService"
#define CLASS_DATA		"GlueServiceData"
#define CLASS_KEY		"GlueKey"
#define CLASS_SA		"GlueSA"

/* LDAP filter definitions */

// for gfal

#define QUERY_RLS "(&(GlueServiceType=edg-*)(|(GlueServiceAccessControlBaseRule=%s)(GlueServiceAccessControlRule=%s)))"
#define QUERY_TYPE_ENDPOINTS	"(|(GlueSEUniqueID=%s)(&(GlueServiceType=srm*)(GlueServiceEndpoint=*://%s:%s*)))"

#define QUERY_LFC_ENDPOINT 	"( &( GlueServiceType=lcg-file-catalog ) %s ) "

#define QUERY_SE_STATUS 	"(&(GlueSEStatus=*Production*)(&(GlueSEUniqueID=*%s*)))"

#define QUERY_PROTO_PORT_TYPE "(&(GlueSEAccessProtocolType=*)(GlueChunkKey=GlueSEUniqueID=%s))"

#define QUERY_CE_AP_ATNM 	"(" ATTR_CE_AP_ATNM	"=%s)"

#define QUERY_SA_PATH	"(&(GlueSALocalID=%s)(" ATTR_CHUNK "=" ATTR_SE "=%s))"

#define QUERY_NAME	"(&" \
					"(" ATTR_OC "=" CLASS_SERVICE ")" \
					"(|(" ATTR_NAME "=%s)" \
					"(" ATTR_URI "=%s))" \
				    ")"

#define QUERY_DATA	"(" ATTR_OC "=" CLASS_DATA ")"

#define QUERY_DATAKEY "(&" \
					    QUERY_DATA \
					  "(" ATTR_DATAKEY "=%s)" \
					  ")"

/* Query fragments for listServicesByData() */
#define QUERY_OC_SERVICE "(" ATTR_OC "=" CLASS_SERVICE ")"
#define QUERY_OC_SA		 "(" ATTR_OC "=" CLASS_SA ")"
#define QUERY_SITE		 "(&" \
					"(" ATTR_OC "=" CLASS_KEY ")" \
					"(" ATTR_LINK "=" ATTR_SITE "=%s)" \
				    ")"
#define QUERY_TYPE "(" ATTR_TYPE "=%s)"

#define QUERY_SRM_TYPE	"(|" \
					"(" ATTR_TYPE "=SRM)" \
					"(" ATTR_TYPE "=srm_v1)" \
				    ")"


#define QUERY_GSIFTP_TYPE	"(|" \
					"(" ATTR_TYPE "=GSIFTP)" \
					"(" ATTR_TYPE "=GSIFTP)" \
				    ")"

#define QUERY_HOST "(|(" ATTR_ENDPOINT "=*%s*)" \
                   "(" ATTR_URI  "=*%s*))"


/*
#define VAL1 "GlueSEControlProtocolEndpoint"
#define VAL2 "GlueServiceEndpoint"

#define QUERY_SE "(" VAL1 "=" VAL2 "=*%s*)"
 */
#define QUERY_SE "(" ATTR_CHUNK "=" ATTR_SE "=%s)"


/* Query expressing "No VO attribute" */
#define QUERY_VO_ANY "(!(" ATTR_VO "=*))"

#define QUERY_VO_PRE  "(|" QUERY_VO_ANY
#define QUERY_VO      "(" ATTR_VO "=%s)"
#define QUERY_VO_POST ")"

/* XXX Possible values? */
#define QUERY_AVAILABLE	"(" ATTR_STATUS "=running)"

/* LDAP query timeout */
static int QUERY_TIMEOUT = 60;

/**********************************************************************
 * Global variables
 */
static char *SD_EMPTY_MESSAGE = ""; /* default error message */
static char *SD_NOMEM_MESSAGE = "Out of memory"; /* error message when there were not enough*/

/* Cached connection handle */
static LDAP *ldcache = NULL;
/* Use count */
static int ldcnt = 0;

/* Data key object attributes */

static char *data_attrs[] ={ATTR_DATAKEY, ATTR_DATAVALUE, NULL};

static char *all_attrs[] ={LDAP_ALL_USER_ATTRIBUTES, NULL};

static char *no_attrs[] ={LDAP_NO_ATTRS, NULL};

static char *link_attrs[] ={ATTR_LINK, NULL};

static char *service_attrs[] ={ATTR_NAME, ATTR_TYPE, ATTR_ENDPOINT, ATTR_VERSION, NULL};

static char *site_attrs[] ={ATTR_LINK, NULL};

static char *sa_attrs[] ={ATTR_SA_ROOT, ATTR_SA_ACBR, NULL};

static char *wsdl_attrs[] ={ATTR_WSDL, NULL};

static char *sa_path_attrs[] ={ATTR_SA_ROOT, ATTR_SA_PATH_ATNM, NULL};

static char *lfc_endpoint_attrs[] ={ATTR_ENDPOINT, NULL};

static char *ce_ap_attrs[] ={ATTR_CE_AP_ATNM, NULL};

static char *seap_attrs[] ={ATTR_PROTO_PORT, ATTR_PROTO_TYPE, NULL};


/**********************************************************************
 * Utility functions
 */


char *vo_global = NULL;
char **fqan_global = NULL;
int fqan_size_global = 0;

void set_gfal_vo(char *vo) {
    vo_global = vo;
}

void set_gfal_fqan(char **fqan, int fqan_size) {
    fqan_global = fqan;
    fqan_size_global = fqan_size;
}

static int get_timeout() {
    char *bdii_env = getenv("LCG_GFAL_BDII_TIMEOUT");
    if (bdii_env != NULL) {
        return atoi(bdii_env);
    }
    return QUERY_TIMEOUT;
}

/*for internal usage*/
void sd_bdii_freeService(SDService *service);
void sd_bdii_freeServiceList(SDServiceList *list);
void sd_bdii_freeServiceDataList(SDServiceDataList *list);
void sd_bdii_freeServiceDetails(SDServiceDetails *details);
void sd_bdii_freeVOList(SDVOList *vos);
void sd_bdii_freeServiceDetailsList(SDServiceDetailsList *servicedetailsList);
void SD_freeVOList(SDVOList *vos);

void SD_setException(SDException *exception, int status, const char *reason) {
    if (!exception) {
        return;
    }

    exception->status = status;
    if (status == SDStatus_SUCCESS) {
        exception->reason = NULL;
    } else {
        if (reason != NULL) {
            exception->reason = (char *) malloc(strlen(reason) + 1);
            if (exception->reason != NULL) {
                strcpy(exception->reason, reason);
            } else {
                exception->reason = SD_NOMEM_MESSAGE;
            }
        } else {
            exception->reason = SD_EMPTY_MESSAGE;
        }
    }
}

/* Return an LDAP connection to the BDII server */
static LDAP *get_connection(SDException *exception) {

    char *lcg_infosys=NULL;
    int ret;
    

    
    if (ldcache)
    {

	ldcnt++;
	return ldcache;
    }
   

    lcg_infosys = getenv("LCG_GFAL_INFOSYS");
    if (!lcg_infosys) {

        SD_setException(exception, SDStatus_FAILURE, "LCG_GFAL_INFOSYS is not set");
        return NULL;
    }

    /* ldap_initialize() can handle the "url://host:port" syntax just fine */
    /*It requires the full uri to be specified*/
    char url[255];
    sprintf(url, "ldap://%s", lcg_infosys);
    ret = ldap_initialize(&ldcache, url);
    if (ret != LDAP_SUCCESS) {

        char buf[256];
        snprintf(buf, sizeof (buf), "ldap_initialize() failed: %s", strerror(errno));
        SD_setException(exception, SDStatus_FAILURE, buf);
        return NULL;
    }
     

    /* REMOVE THE COMMENS FOLLOWING TO ENABLE BER AND LDAP DEBUGGING 
    int debug = 7;
        
    if ( ber_set_option( NULL, LBER_OPT_DEBUG_LEVEL, &debug ) != LBER_OPT_SUCCESS ) {
  	fprintf( stderr, "Could not set LBER_OPT_DEBUG_LEVEL %d\n", debug );
    }
    
    
    if ( ldap_set_option( NULL, LDAP_OPT_DEBUG_LEVEL, &debug ) != LDAP_OPT_SUCCESS ) {
  	fprintf( stderr, "Could not set LDAP_OPT_DEBUG_LEVEL %d\n", debug );
    }    
   */   
        
    struct berval cred;
    cred.bv_val = NULL;
    cred.bv_len = 0;


    ret = ldap_sasl_bind_s(ldcache, NULL, LDAP_SASL_SIMPLE, &cred, NULL, NULL, NULL);
    if (ret != LDAP_SUCCESS) {  

        char buf[256];
        snprintf(buf, sizeof (buf), "ldap_sasl_bind_s() failed: %s", strerror(errno));
        SD_setException(exception, SDStatus_FAILURE, buf);
        ldap_unbind_ext_s(ldcache, NULL, NULL);
        ldcache = NULL;
        return NULL;
    }
    


    ldcnt++;
    return ldcache;
}

static void close_connection(LDAP *ld) {


    if (--ldcnt){

	return;
    }   
    
    if(ld){

	    ldap_unbind_ext_s(ld, NULL, NULL);
	    ldcache = NULL;
    }


}

void report_query_error(const char *fmt, LDAP *ld, LDAPMessage *msg, int err,
        SDException *exception) {
    char *error=NULL;
    int ret;
    if (NULL != msg && ld != NULL) {
        ret = ldap_parse_result(ld, msg, NULL, NULL, NULL, NULL, NULL, TRUE);
    } else {
        ret = err;
    }

    error = g_strdup_printf(fmt, ldap_err2string(ret));

    if (error != NULL)
        fprintf(stderr, "Error: %s\n", error);

    SD_setException(exception, SDStatus_FAILURE, error);

    g_free(error);
}

#if 0

void report_ldap_error(const char *fmt, LDAP *ld, SDException *exception) {
    char *error;
    int ret;

    ldap_get_option(ld, LDAP_OPT_ERROR_NUMBER, &ret);
    error = g_strdup_printf(fmt, ldap_err2string(ret));
    SD_setException(exception, SDStatus_FAILURE, error);
    g_free(error);
}
#endif

/* Get the value of a single-valued attribute */
char *get_single_value(LDAP *ld, LDAPMessage *msg, const char *attr) {

    struct berval **values=NULL;
    char *res=NULL;

    values = ldap_get_values_len(ld, msg, attr);
    if (!values) {
        return NULL;
    }

    if (values && values[0] && values[0]->bv_val) {
        res = g_strdup(values[0]->bv_val);
        ber_bvecfree(values);
    }
    return res;
}

int parse_service_attrs(SDService *service, LDAP *ld, LDAPMessage *msg,
        SDException *exception) {

    service->name = get_single_value(ld, msg, ATTR_NAME);
    service->type = get_single_value(ld, msg, ATTR_TYPE);
    service->endpoint = get_single_value(ld, msg, ATTR_ENDPOINT);
    service->version = get_single_value(ld, msg, ATTR_VERSION);


    if (NULL == service->version) {
        char *major = get_single_value(ld, msg, ATTR_VERSION_MAJOR);
        char *minor = get_single_value(ld, msg, ATTR_VERSION_MINOR);
        char *patch = get_single_value(ld, msg, ATTR_VERSION_PATCH);
        service->version = (char *) malloc(20 * sizeof (char));
        if (major && minor && patch) {

            sprintf(service->version, "%s.%s.%s", major, minor, patch);
            g_free(major);
            g_free(minor);
            g_free(patch);
        } else if (major) {

            sprintf(service->version, "%s", major);
            g_free(major);
        } else {
            sprintf(service->version, "N/A");
        }
    }

    if (NULL == service->endpoint) {
        service->endpoint = get_single_value(ld, msg, ATTR_URI);
    }

    if (NULL == service->name) {
        service->name = get_single_value(ld, msg, ATTR_NM);
        if (NULL == service->name) {
            service->name = get_single_value(ld, msg, ATTR_URI);
        }
    }

    if (!service->name || !service->type || !service->endpoint) {
        char msg[1000];
        sprintf(
                msg,
                "Incomplete service definition: ServiceName:%s ServiceType:%s ServiceEndpoint:%s ServiceVersion:%s",
                !service->name ? "N/A" : service->name, !service->type ? "N/A"
                : service->type, !service->endpoint ? "N/A"
                : service->endpoint, !service->version ? "N/A"
                : service->version);
        SD_setException(exception, SDStatus_FAILURE, msg);
        return -1;
    }

    // the SRM fix (you can start laughing now)
    if (strcmp(service->type, "srm_v1") == 0) {
        strcpy(service->type, "SRM");
    }
    
    return 0;
}

int match_vos(LDAP *ld, LDAPMessage *msg, const SDVOList *vos) {

    struct berval **values=NULL;
    int i, j;

    // try the VO as we know it (owner)
    values = ldap_get_values_len(ld, msg, ATTR_VO);
    if (!values) {
        // try the VO as it is sometimes used (accessrule)
        values = ldap_get_values_len(ld, msg, ATTR_ACCESSRULE);
    }

    // if no VO is given at all, return TRUE, assume that the service is for all VOs
    if (!values)
        return TRUE;

    for (i = 0; i < vos->numNames; i++) {
        for (j = 0; j < ldap_count_values_len(values); j++) {
            if (values && values[j] && values[j]->bv_val) {
                if (0 == strcasecmp(vos->names[i], values[j]->bv_val)) {
                    ber_bvecfree(values);
                    return TRUE;
                }
            }
        }
    }

    if(values)
       ber_bvecfree(values);
    return FALSE;
}

/**
 * Utility function, connects and searched the ldap server
 * Param	filter/query, attr to be searched, pointer to an ldap server
 * Return	LDAPMessage pointer
 */

static int connect_search_ldap(char *query, char *attr[], LDAP *ld,
        SDException *exception, LDAPMessage **reply) {

    struct timeval timeout;
    int ret;

    if (!exception)
        exception = g_newa(SDException, 1);

    timeout.tv_sec = get_timeout();
    timeout.tv_usec = 0;


    ret = ldap_search_ext_s(ld, sbasedn, LDAP_SCOPE_SUBTREE, query, attr, 0, NULL, NULL, &timeout, 0, reply);
    if(query)
    	g_free(query);

        return ret;
}

/**
 * @description	fills a structure with attr as: name, type, endpoint, version */
int fill_service(const char *name, SDService *service, SDException *exception) {

    LDAPMessage *reply=NULL;
    char *query=NULL;
    LDAP *ld=NULL;
    int ret=0;

    if (!exception)
        exception = g_newa(SDException, 1);

    query = g_strdup_printf(QUERY_NAME, name, name);

    /* returns an LDAP reply */
    if ((ld = get_connection(exception)) == NULL) {
        SD_setException(exception, SDStatus_FAILURE, "Cannot connect to BDII");
        return -1;
    }
    ret = connect_search_ldap(query, all_attrs, ld, exception, &reply);
    if (ret != LDAP_SUCCESS) {
        SD_setException(exception, SDStatus_FAILURE, "fill_service query error");
        /* clean up */
        if(reply){
             ldap_msgfree(reply);
	}
	close_connection(ld);
        return -1;
    }

    /* copy attr from reply to the SDService Structure */
    ret = parse_service_attrs(service, ld, reply, exception);

    /* clean up */
    if(reply){
        ldap_msgfree(reply);
    }
        close_connection(ld);

    return ret;
}

int add_associated_service(SDServiceDetails *details, const char *name,
        SDException *exception) {

    SDService *service, **tmp;
    int ret;

    if (!exception)
        exception = g_newa(SDException, 1);

    service = g_new0(SDService, 1);
    ret = fill_service(name, service, exception);
    if (ret) {
        /* We do not want to report an error if the pointed service does not exist */
        sd_bdii_freeService(service);
        return 0;
    }

    if (!details->associatedServices)
        details->associatedServices = g_new0(SDServiceList, 1);

    tmp = g_renew(SDService *, details->associatedServices->services,
            details->associatedServices->numServices + 1);
    tmp[details->associatedServices->numServices++] = service;
    details->associatedServices->services = tmp;
    return 0;
}

int add_vo(SDServiceDetails *details, const char *vo) {

    if (!details->vos)
        details->vos = g_new0(SDVOList, 1);

    details->vos->names
            = g_renew(char *, details->vos->names, details->vos->numNames + 1);
    details->vos->names[details->vos->numNames++] = g_strdup(vo);
    return 0;
}

/* Add a service to a SDServiceList
 * add a structure like SDService to a list. */
int add_to_list(SDServiceList *list, SDService *service, SDException *ex) {

    SDService **tmp=NULL;

    if (!ex)
        ex = g_newa(SDException, 1);

    tmp = g_renew(SDService *, list->services, list->numServices + 1);
    if (!tmp) {
        SD_setException(ex, SDStatus_FAILURE, "Out of memory");
        return -1;
    }

    tmp[list->numServices++] = service;
    list->services = tmp;
    return 0;
}

int parse_detail_attrs(SDServiceDetails *details, LDAP *ld, LDAPMessage *msg,
        SDException *exception) {

    struct berval **values=NULL;
    int i, ret;

    SDService *service=NULL;
    service = g_new0(SDService, 1);

    if (!exception)
        exception = g_newa(SDException, 1);

    ret = parse_service_attrs(service, ld, msg, exception);
    if (ret) {        
        return ret;
    }

    details->name = service->name;
    details->type = service->type;
    details->endpoint = service->endpoint;
    details->version = service->version;
    details->wsdl = get_single_value(ld, msg, ATTR_WSDL);


    /* details->administration */
    /* ->site and ->associatedServices comes from ATTR_LINK */

    values = ldap_get_values_len(ld, msg, ATTR_LINK);
    ret = 0;
    for (i = 0; values && values[i] && values[i]->bv_val; i++) {

        char *name, *val;

        name = g_strdup(values[i]->bv_val);
        val = strchr(name, '=');
        if (!val) {

            g_free(name);
            continue;
        }

        *val++ = '\0';

        if (!strcasecmp(name, ATTR_SITE)) {

            g_free(details->site);
            details->site = g_strdup(val);
        } else if (!strcasecmp(name, ATTR_NAME)) {

            ret = add_associated_service(details, val, exception);
            if (ret)
                break;
        }

        g_free(name);

    }


    ber_bvecfree(values);
    if (ret) {
        return ret;
    }

    if (!details->site) {
        // it can be that the site is attached differently (old LCG model)
        details->site = get_single_value(ld, msg, ATTR_HOSTINGORG);
    }

    values = ldap_get_values_len(ld, msg, ATTR_ACCESSRULE);
    if (!values) {
        // try the VO as it should be used
        values = ldap_get_values_len(ld, msg, ATTR_VO);
    }
    for (i = 0; values && values[i] && values[i]->bv_val; i++) {


        ret = add_vo(details, values[i]->bv_val);
        if (ret) {
            break;
        }
    }
    ber_bvecfree(values);

    if (ret)
    {
    
        return ret;
    }
    SD_setException(exception, SDStatus_SUCCESS, "Success");

    return 0;
}

char *get_host(const char *endpoint) {

    char * ret = g_strdup(endpoint);

    char *start = strstr(ret, "://");
    if (NULL != start) {
        ret = start + 3;
    }

    char *end = strchr(ret, ':');
    if (NULL != end) {
        *end = '\0';
    } else {
        end = strchr(ret, '/');
        if (NULL != end)
            *end = '\0';
    }

    return ret;

}

void add_sa_root(SDServiceDataList *list, const char *endpoint,
        SDException *exception) {

    LDAPMessage *reply=NULL;
    LDAPMessage *msg=NULL;
    LDAP *ld=NULL;
    GString *query=NULL;
    int ret;

    if (!exception)
        exception = g_newa(SDException, 1);

    char *host = get_host(endpoint);

    query = g_string_new("(&" QUERY_OC_SA);
    g_string_append_printf(query, QUERY_SE, host);
    g_string_append(query, ")");

    if ((ld = get_connection(exception)) == NULL) {
        SD_setException(exception, SDStatus_FAILURE,
                "add_sa_root: Cannot connect to BDII");
        goto out;
    }
    /* returns an LDAP reply */
    ret = connect_search_ldap(query->str, sa_attrs, ld, exception, &reply);
    if (ret != LDAP_SUCCESS) {
        SD_setException(exception, SDStatus_FAILURE,
                "add_sa_root: connect_search_ldap failed");
        goto out;
    }

    if (ldap_count_entries(ld, reply) == 0) {

        goto out;
    }


    /* iterate the entries */
    for (msg = ldap_first_entry(ld, reply); msg; msg = ldap_next_entry(ld, msg)) {

        char *vo = get_single_value(ld, msg, ATTR_SA_ACBR);
        char *root = get_single_value(ld, msg, ATTR_SA_ROOT);

        if (!vo || !root)
            continue;

        char *value = strchr(root, ':');
        char *key = g_strdup_printf("%s:SEMountPoint", vo);
        g_free(vo);

        list->items = g_renew(SDServiceData, list->items, list->numItems + 1);
        list->items[list->numItems].key = key;

        if (NULL != value) {
            list->items[list->numItems].value = ++value;
        } else {
            list->items[list->numItems].value = root;
        }
        ++list->numItems;
	
    }

out:
    if(reply){
      ldap_msgfree(reply);
    }
      close_connection(ld);    
    return;

}

SDServiceDataList *query_service_data(LDAP *ld, LDAPMessage *imsg,
        const char *type, const char *endpoint, SDException *exception) {
    char *dn;
    SDServiceDataList *list = NULL;
    LDAPMessage *reply=NULL;
    LDAPMessage *msg=NULL;
    struct timeval timeout;
    int ret;

    if (!exception)
        exception = g_newa(SDException, 1);

    dn = ldap_get_dn(ld, imsg);
    if (!dn) {
        SD_setException(exception, SDStatus_FAILURE,
                "cannot get DN");
        return NULL;
    }


    timeout.tv_sec = get_timeout();
    timeout.tv_usec = 0;


    ret = ldap_search_ext_s(ld, dn, LDAP_SCOPE_SUBTREE, QUERY_DATA, data_attrs, 0, NULL, NULL,
            &timeout, 0, &reply);
    ldap_memfree(dn);

    if (ret != LDAP_SUCCESS) {
        report_query_error("LDAP search failed:  %s", ld, reply, ret, exception);
        goto error;
    }

    list = g_new0(SDServiceDataList, 1);

    for (msg = ldap_first_entry(ld, reply); msg; msg = ldap_next_entry(ld, msg)) {

        char *key, *value;

        key = get_single_value(ld, msg, ATTR_DATAKEY);
        value = get_single_value(ld, msg, ATTR_DATAVALUE);


        if (!key || !value) {

            g_free(key);
            g_free(value);
            continue;
        }

        list->items = g_renew(SDServiceData, list->items, list->numItems + 1);
        list->items[list->numItems].key = key;
        list->items[list->numItems].value = value;
        ++list->numItems;
    }


   if(ret > 0)
    ldap_msgfree(reply);

    // SPECIAL case for the SE/SRM case
    if (strncasecmp(type, "srm", 3) == 0) {

        add_sa_root(list, endpoint, exception);
    }

    return list;

error:
    sd_bdii_freeServiceDataList(list);
    if(ret > 0)
      ldap_msgfree(reply);
    return NULL;
}

int match_data(const SDServiceDataList *list,
        const SDServiceDataList *requirements) {

    int i, j, gotit;

    /* Check that the service has all the specified key/value
     * pairs */
    for (i = 0; i < requirements->numItems; i++) {
        gotit = 0;

        for (j = 0; list && j < list->numItems; j++) {
            if (strcmp(requirements->items[i].key, list->items[j].key))
                continue;
            if (!strcmp(requirements->items[i].value, list->items[j].value)) {
                gotit = 1;
                break;
            }
        }

        if (!gotit)
            return FALSE;
    }

    return TRUE;
}

/* Returns TRUE if the service matches the given constraints, FALSE otherwise */
int match_service(SDServiceDetails *details, const char *type,
        const char *site, const SDVOList *vos) {

    int i, j, gotit;

    if (type && strcmp(details->type, type))
        return FALSE;
    if (site && details->site && strcmp(details->site, site))
        return FALSE;

    /* <volist> is not a required parameter. Treat an missing <volist>
     * as a "match all", an empty <volist> as a "match none" */
    if (vos && details->vos) {
        gotit = 0;

        /* A service that does not belong to any VOs matches the
         * empty list */
        if (!vos->numNames && !details->vos->numNames)
            gotit = 1;

        /* Check that at the intersection of vos and details->vos
         * is not empty */
        for (i = 0; i < vos->numNames && !gotit; i++) {
            for (j = 0; details->vos && j < details->vos->numNames; j++) {
                if (!strcmp(vos->names[i], details->vos->names[j])) {
                    gotit = 1;
                    break;
                }
            }
        }

        if (!gotit)
            return FALSE;
    }

    return TRUE;
}

/* Convert SDServiceDetails to SDService */
SDService *details_to_service(const SDServiceDetails *details) {

    SDService *service=NULL;

    service = g_new0(SDService, 1);
    service->name = g_strdup(details->name);
    service->type = g_strdup(details->type);
    service->endpoint = g_strdup(details->endpoint);
    service->version = g_strdup(details->version);
    return service;
}

int get_details(const char *q, const char *name, LDAP *ld, LDAPMessage **reply,
        SDException *exception) {
    char *query=NULL;
    struct timeval timeout;
    int ret;

    query = g_strdup_printf(q, name, name);

    timeout.tv_sec = get_timeout();
    timeout.tv_usec = 0;


    ret = ldap_search_ext_s(ld, sbasedn, LDAP_SCOPE_SUBTREE, query, all_attrs, 0, NULL, NULL,
            &timeout, 0, reply);
    g_free(query);
    if (ret != LDAP_SUCCESS) {
        report_query_error("LDAP search failed: %s", ld, *reply, ret, exception);
        return -1;
    }
    return 0;
}

SDServiceList *do_list_query(const SDServiceDataList *data,
        const SDVOList *vos, LDAP *ld, GString *query, SDException *exception) {
    LDAPMessage *reply=NULL;
    LDAPMessage *msg;
    SDServiceList *list;
    int ret;

    if (!exception)
        exception = g_newa(SDException, 1);
   
    /* returns an LDAP reply */
    ret = connect_search_ldap(query->str, all_attrs, ld, exception, &reply);
    if (ret != LDAP_SUCCESS) {
        report_query_error("LDAP search failed: %s", ld, reply, ret, exception);
        list = NULL;
        goto out;
    }

    if (ldap_count_entries(ld, reply) == 0) {
        SD_setException(exception, SDStatus_FAILURE, "No entries found");
        list = NULL;
        goto out;
    }


    list = g_new0(SDServiceList, 1);
    for (msg = ldap_first_entry(ld, reply); msg; msg = ldap_next_entry(ld, msg)) {

        SDService *service=NULL;

        service = g_new0(SDService, 1);
        ret = parse_service_attrs(service, ld, msg, exception);
        if (ret) {

            sd_bdii_freeService(service);
            sd_bdii_freeServiceList(list);
            list = NULL;
            goto out;
        }

        if (data) {


            SDServiceDataList *datalist=NULL;

            SD_setException(exception, SDStatus_SUCCESS, "Found");
            datalist = query_service_data(ld, msg, service->type,
                    service->endpoint, exception);
            if (exception->status != SDStatus_SUCCESS)
                continue;

            /* Filter out non-matching entries */
            if (!match_data(datalist, data))
                continue;
        }

        if (vos) {

            /* Filter out entries not matching the requested VOs */
            if (!match_vos(ld, msg, vos))
                continue;
        }

        ret = add_to_list(list, service, exception);
        if (ret) {
            sd_bdii_freeService(service);
            sd_bdii_freeServiceList(list);
            list = NULL;
            goto out;
        }

    }

    SD_setException(exception, SDStatus_SUCCESS, "Success");
out:
    if(reply){
      ldap_msgfree(reply);
    }
    close_connection(ld);
    return list;
}

SDServiceDetails *SD_getServiceDetails(const char *name,
        SDException *exception) {
    SDServiceDetails *details = NULL;
    LDAPMessage *reply=NULL;
    LDAP *ld=NULL;
    int ret;

    ld = get_connection(exception);
    if (!ld) {
        return NULL;
    }

    if (!exception)
        exception = g_newa(SDException, 1);

    ret = get_details(QUERY_NAME, name, ld, &reply, exception);
    if (ret < 0) {
        goto error;
    }

    // nothing ?
    if (ldap_count_entries(ld, reply) == 0) {
        SD_setException(exception, SDStatus_FAILURE, "No entry found");
        goto error;
    }


    details = g_new0(SDServiceDetails, 1);
    ret = parse_detail_attrs(details, ld, reply, exception);


    if (ret) {
        goto error;
    }
    SD_setException(exception, SDStatus_SUCCESS, "Success");

    details->data = query_service_data(ld, reply, details->type, details->endpoint, exception);
    if (exception->status != SDStatus_SUCCESS) {
        goto error;
    }

   

out:
    if(reply){
       ldap_msgfree(reply);
    }
      close_connection(ld);    
    return details;

error:
    if (details)
        sd_bdii_freeServiceDetails(details);
    details = NULL;
    goto out;


}

SDServiceList *SD_listAssociatedServices(const char *name,
        const char *type, const char *site, const SDVOList *vos,
        SDException *exception) {
    char *query=NULL;
    struct berval **values=NULL;
    SDServiceList *list=NULL;
    SDServiceDetailsList *detailslist=NULL;
    LDAPMessage *reply=NULL;
    int ret, i;
    LDAP *ld=NULL;

    if (!exception)
        exception = g_newa(SDException, 1);

    query = g_strdup_printf(QUERY_NAME, name, name);

    /* returns an LDAP reply */
    if ((ld = get_connection(exception)) == NULL) {
        return NULL;
    }

    /* returns an LDAP reply */
    ret = connect_search_ldap(query, link_attrs, ld, exception, &reply);
    if (ret != LDAP_SUCCESS) {
         report_query_error("LDAP search failed: %s", ld, reply, ret, exception);
	 if(reply){
            ldap_msgfree(reply);
	   }
           close_connection(ld);	   
         return NULL;
    }


    list = g_new0(SDServiceList, 1);
    detailslist = g_new0(SDServiceDetailsList, 1);
    if (!list || !detailslist) {
        SD_setException(exception, SDStatus_FAILURE, "Out of memory");
        return NULL;
    }

    values = ldap_get_values_len(ld, reply, ATTR_LINK);
    for (i = 0; values && values[i] && values[i]->bv_val; i++) {
        SDServiceDetails *details;
        SDService *service;

        /* We are only interested in service links */
        if (strncasecmp(values[i]->bv_val, ATTR_NAME, strlen(ATTR_NAME))
                || values[i]->bv_val[strlen(ATTR_NAME)] != '=')
            continue;

        /* A non-existant associated service is not an error */
        details = SD_getServiceDetails(values[i]->bv_val + strlen(
                ATTR_NAME) + 1, exception);
        if (!details)
            continue;

        /* Check if the service matches the requirements */
        if (!match_service(details, type, site, vos)) {
            sd_bdii_freeServiceDetails(details);
            continue;
        }

        service = details_to_service(details);
        ret = add_to_list(list, service, exception);
        if (ret) {
            sd_bdii_freeService(service);
            sd_bdii_freeServiceList(list);
            sd_bdii_freeServiceDetails(details);
            list = NULL;
            goto out;
        }
    }

    if (!list->numServices) {
        char buf[256];
        sd_bdii_freeServiceList(list);
        snprintf(buf, 255, "No services of type %s were found", type);
        SD_setException(exception, SDStatus_FAILURE, buf);
        list = NULL;
        goto out;
    }

    SD_setException(exception, SDStatus_SUCCESS, "Success");

out:
    ber_bvecfree(values);
    if(reply){
       ldap_msgfree(reply);
    }
       close_connection(ld);    
    return list;


}

SDServiceList *SD_listServicesByHost(const char *type, const char *host,
        const SDVOList *vos, SDException *exception) {

    GString *query=NULL;
    LDAP *ld=NULL;

    if (!exception)
        exception = g_newa(SDException, 1);

    if (host == NULL){
        return SD_listServicesByData(NULL, type, NULL, vos, exception);
	}

    /* returns an LDAP reply */
    if ((ld = get_connection(exception)) == NULL) {
        return NULL;
    }

    /* Set up a general LDAP error message */
    SD_setException(exception, SDStatus_FAILURE, "LDAP query failed");


    /* GString is _very_ handy here :-) */
    query = g_string_new("(&" QUERY_OC_SERVICE);

    if (type) {
        // this is the SRM hack so that both old and new types are looked for
        if (strncasecmp(type, "srm", 3) == 0) {
	   
            g_string_append(query, QUERY_SRM_TYPE);
        } else if (strncasecmp(type, "gsiftp", 3) == 0) {
		
            g_string_append(query, QUERY_GSIFTP_TYPE);
        }
        else {
		
            g_string_append_printf(query, QUERY_TYPE, type);
        }
    }

    g_string_append_printf(query, QUERY_HOST, host, host);

    /* closing the and-ed subqueries */
    g_string_append(query, ")");

    return do_list_query(NULL, vos, ld, query, exception);

}

SDServiceList *SD_listServicesByData(const SDServiceDataList *data,
        const char *type, const char *site, const SDVOList *vos,
        SDException *exception) {
    /* type, site and vos can be expressed as an LDAP filter. Data
     * key-value pairs however appear as child nodes in BDII so we have
     * to check them manually. */
    GString *query=NULL;
    LDAP *ld=NULL;

    if (!exception)
        exception = g_newa(SDException, 1);


    ld = get_connection(exception);
    if (!ld) {
        return NULL;
    }


    /* GString is _very_ handy here :-) */
    if (type || site)
        query = g_string_new("(&" QUERY_OC_SERVICE);
    else
        query = g_string_new(QUERY_OC_SERVICE);


    if (type) {
        // this is the SRM hack so that both old and new types are looked for
        if (strncasecmp(type, "srm", 3) == 0) {
            g_string_append(query, QUERY_SRM_TYPE);
        } else if (strncasecmp(type, "gsiftp", 3) == 0) {
            g_string_append(query, QUERY_GSIFTP_TYPE);
        }
        else {
            g_string_append_printf(query, QUERY_TYPE, type);
        }
    }
    if (site)
        g_string_append_printf(query, QUERY_SITE, site);

    /* closing the and-ed subqueries */
    if (type || site)
        g_string_append(query, ")");


    return do_list_query(data, vos, ld, query, exception);
}

char *dupstr(const char *from) {
    char *to;

    to = (char *) malloc(strlen(from) + 1);
    if (to != NULL) strcpy(to, from);
    return to;
}

SDVOList *checkVOEnv(void) {
    SDVOList *volist;
    const char *vo;

    vo = getenv(SDEnv_VO);
    if (!vo)
        return NULL;

    if (strlen(vo) == 0)
        return NULL;

    volist = (SDVOList *) calloc(1, sizeof (SDVOList));
    if (!volist)
        return NULL;

    volist->names = (char **) malloc(sizeof (char *));
    if (!volist->names) {
        free(volist);
        return NULL;
    }

    volist->numNames = 1;
    volist->names[0] = dupstr(vo);
    if (!volist->names[0]) {
        free(volist->names);
        free(volist);
        return NULL;
    }
    return volist;
}

SDServiceList *SD_listServices(const char *type, const char *site, const SDVOList *vos, SDException *exception) {
    SDServiceList *result = NULL;
    SDVOList *tmpvos=NULL;

    if (!exception)
        exception = g_newa(SDException, 1);

    if (!site) {
        site = getenv(SDEnv_SITE);
        /* treat empty string as if it does not exist */
        if (site && (strlen(site) == 0)) site = NULL;
    }
    tmpvos = NULL;
    /* Let the user override the VO if the caller did not specify
     * one explicitely */
    if (!vos || !vos->numNames)
        tmpvos = checkVOEnv();


    result = SD_listServicesByData(NULL, type, site, tmpvos ? tmpvos : vos, exception);
    if (result == NULL) {
        return NULL;
    }


    if (tmpvos)
        SD_freeVOList(tmpvos);


    SD_setException(exception, SDStatus_SUCCESS, "Success");
    return result;
}

SDService *SD_getService(const char *name, SDException *exception) {

    SDService *service = NULL;
    GString *query = NULL;
    LDAP *ld=NULL;

    ld = get_connection(exception);
    if (!ld) {
        return NULL;
    }

    /* do_list_query returns the error code in the exception */
    if (!exception)
        exception = g_newa(SDException, 1);

    SD_setException(exception, SDStatus_FAILURE, "LDAP query failed");

    /* GString is _very_ handy here :-) */
    if (!name) {
        SD_setException(exception, SDStatus_FAILURE,
                "Name not given in Service search");
        return NULL;
    }

    query = g_string_sized_new(200);
    g_string_printf(query, QUERY_NAME, name, name);

    SDServiceList *list = do_list_query(NULL, NULL, ld, query, exception);
    if (list == NULL) {
        return NULL;
    }
    if (list->numServices < 1) {
        return NULL;
    }

    service = list->services[0];
    g_free(list->services);
    g_free(list);


    /*SD_setException(exception, SDStatus_SUCCESS, "Success");*/
    return service;
}

SDServiceDataList *SD_getServiceData(const char *name,
        SDException *exception) {

    SDServiceDataList *list = NULL;

    LDAPMessage *reply=NULL;
    char *query=NULL;
    LDAP *ld=NULL;
    int ret;

    /* query_service_data() returns the error code in the exception */
    if (!exception)
        exception = g_newa(SDException, 1);

    query = g_strdup_printf(QUERY_NAME, name, name);

    /* returns an LDAP reply */
    if ((ld = get_connection(exception)) == NULL) {
        return NULL;
    }

    ret = connect_search_ldap(query, service_attrs, ld, exception, &reply);
    if (ret != LDAP_SUCCESS) {
        report_query_error("LDAP search failed: %s", ld, reply, ret, exception);
        list = NULL;
        goto out;
    }



    SDService *service;

    service = g_new0(SDService, 1);
    ret = parse_service_attrs(service, ld, reply, exception);
    if (ret) {
        sd_bdii_freeService(service);
        list = NULL;
        goto out;
    }

    SD_setException(exception, SDStatus_SUCCESS, "Success");
    list = query_service_data(ld, reply, service->type, service->endpoint,
            exception);

out:
    if(reply){
        ldap_msgfree(reply);
        
    }
    close_connection(ld);
    return list;
}

char *SD_getServiceSite(const char *name, SDException *exception) {

    char *query, *res = NULL;
    struct berval **values=NULL;
    LDAPMessage *reply=NULL;
    int ret, i;
    LDAP *ld=NULL;

    /* query_service_data() returns the error code in the exception */
    if (!exception)
        exception = g_newa(SDException, 1);


    query = g_strdup_printf(QUERY_NAME, name, name);

    /* returns an LDAP reply */
    if ((ld = get_connection(exception)) == NULL) {
        return NULL;
    }
    ret = connect_search_ldap(query, site_attrs, ld, exception, &reply);
    if (ret != LDAP_SUCCESS) {
        report_query_error("LDAP search failed: %s", ld, reply, ret, exception);
        goto out;
    }
    values = ldap_get_values_len(ld, reply, ATTR_LINK);
    for (i = 0; values && values[i] && values[i]->bv_val; i++) {
        if (strncasecmp(values[i]->bv_val, ATTR_SITE, strlen(ATTR_SITE))
                || values[i]->bv_val[strlen(ATTR_SITE)] != '=')
            continue;
        res = strdup(values[i]->bv_val + strlen(ATTR_SITE) + 1);
    }
    ber_bvecfree(values);

out:
    if(reply){
      ldap_msgfree(reply);
      
    }
    close_connection(ld);
    return res;

}

char *SD_getServiceWSDL(const char *name, SDException *exception) {

    char *query, *res = NULL;
    LDAPMessage *reply=NULL;
    LDAP *ld;
    int ret;

    if (!exception)
        exception = g_newa(SDException, 1);

    query = g_strdup_printf(QUERY_NAME, name, name);

    /* returns an LDAP reply */
    if ((ld = get_connection(exception)) == NULL) {
        return NULL;
    }
    ret = connect_search_ldap(query, wsdl_attrs, ld, exception, &reply);
    if (ret != LDAP_SUCCESS) {
        report_query_error("LDAP search failed: %s", ld, reply, ret, exception);
        goto out;
    }

    res = get_single_value(ld, reply, ATTR_WSDL);
    if (!g_mem_is_system_malloc() && res) {
        char *tmp = strdup(res);
        g_free(res);
        res = tmp;
    }


    SD_setException(exception, SDStatus_SUCCESS, "Success");

out:
    if(reply){
       ldap_msgfree(reply);
       
    }
    close_connection(ld);
    return res;
}

char *SD_getServiceDataItem(const char *name, const char *key,
        SDException *exception) {

    char *query, *dn, *res = NULL;
    struct timeval timeout;
    LDAPMessage *reply=NULL;
    LDAP *ld;
    int ret;

    if (!exception)
        exception = g_newa(SDException, 1);

    query = g_strdup_printf(QUERY_NAME, name, name);

    /* returns an LDAP reply */
    if ((ld = get_connection(exception)) == NULL) {
        return NULL;
    }
    ret = connect_search_ldap(query, no_attrs, ld, exception, &reply);
    if (ret != LDAP_SUCCESS) {
        report_query_error("LDAP search failed: %s", ld, reply, ret, exception);
        goto out;
    }
    dn = ldap_get_dn(ld, reply);
    if (!dn) {
        SD_setException(exception, SDStatus_FAILURE,
                "Failed to determine the DN");
        goto out;
    }

    if(ret > 0)
      ldap_msgfree(reply);

    query = g_strdup_printf(QUERY_DATAKEY, key);
    timeout.tv_sec = get_timeout();
    timeout.tv_usec = 0;


    ret = ldap_search_ext_s(ld, dn, LDAP_SCOPE_SUBTREE, query, data_attrs, 0, NULL, NULL,
            &timeout, 0, &reply);
    g_free(query);
    ldap_memfree(dn);
    if (ret != LDAP_SUCCESS) {
        report_query_error("LDAP search failed: %s", ld, reply, ret, exception);
        goto out;
    }

    res = get_single_value(ld, reply, ATTR_DATAVALUE);
    if (!g_mem_is_system_malloc() && res) {
        char *tmp = strdup(res);
        g_free(res);
        res = tmp;
    }

    SD_setException(exception, SDStatus_SUCCESS, "Success");

out:
    if(reply){
      ldap_msgfree(reply);
      
    }
    close_connection(ld);
    return res;
}

/**********************************************************************
 * Memory management methods
 */

void sd_bdii_freeService(SDService *service) {
    if (!service)
        return;

    g_free(service->name);
    g_free(service->type);
    g_free(service->endpoint);
    g_free(service->version);
    g_free(service);
}

void sd_bdii_freeServiceList(SDServiceList *list) {
    int i;

    if (!list)
        return;

    for (i = 0; i < list->numServices; i++)
        sd_bdii_freeService(list->services[i]);
    g_free(list->services);
    g_free(list);
}

void sd_bdii_freeServiceDataList(SDServiceDataList *list) {
    int i;

    if (!list)
        return;
    for (i = 0; i < list->numItems; i++) {
        g_free(list->items[i].key);
        g_free(list->items[i].value);
    }
    g_free(list->items);
    g_free(list);
}

void sd_bdii_freeServiceDetails(SDServiceDetails *details) {
    if (!details)
        return;

    g_free(details->name);
    g_free(details->type);
    g_free(details->endpoint);
    g_free(details->version);
    g_free(details->site);
    g_free(details->wsdl);
    g_free(details->administration);
    sd_bdii_freeVOList(details->vos);
    sd_bdii_freeServiceList(details->associatedServices);
    sd_bdii_freeServiceDataList(details->data);
    g_free(details);
}

void sd_bdii_freeServiceDetailsList(SDServiceDetailsList *servicedetailsList) {
    int i;
    if (!servicedetailsList)
        return;

    for (i = 0; i < servicedetailsList->numServiceDetails; i++)
        g_free(servicedetailsList->servicedetails[i]);
    g_free(servicedetailsList->servicedetails);
    g_free(servicedetailsList);
}

void sd_bdii_freeVOList(SDVOList *vos) {
    int i;

    if (!vos)
        return;
    for (i = 0; i < vos->numNames; i++)
        g_free(vos->names[i]);
    g_free(vos->names);
    g_free(vos);
}

// *  get from the BDII the CE Accesspoint for a given SE
// *
// *  */

int sd_get_ce_ap(const char *host, char **ce_ap) {
    LDAPMessage *entry=NULL;
    SDException *exception = NULL;
    LDAP *ld;
    LDAPMessage *reply=NULL;
    struct berval **value=NULL;
    char *query=NULL;
    int sav_errno = 0;
    int ret;
    *ce_ap = NULL;
    int rc = 0;

    query = g_strdup_printf(QUERY_CE_AP_ATNM, host);

    if (!exception)
        exception = g_newa(SDException, 1);


    while (!*ce_ap && !sav_errno) {

        /* returns an LDAP connection */
        if ((ld = get_connection(exception)) == NULL) {
           return -1;
        }

        rc = connect_search_ldap(query, ce_ap_attrs, ld, exception, &reply);
        if (rc != LDAP_SUCCESS) {
	    if(rc > 0){
                ldap_msgfree(reply);
	        close_connection(ld);
		}
            sav_errno = errno;
            break;
        }

        if ((entry = ldap_first_entry(ld, reply)) && (value
                = ldap_get_values_len(ld, entry, ATTR_CE_AP_ATNM))) {
            if (value && value[0]->bv_val) {
                if ((*ce_ap = strdup(value[0]->bv_val)) == NULL) {
                    sav_errno = errno ? errno : ENOMEM;
                }
                ber_bvecfree(value);
            }
        } else {
            SD_setException(exception, SDStatus_FAILURE,
                    "No GlueCESEBindCEAccesspoint found.");
        }
        if(rc > 0){
	  ldap_msgfree(reply);
          close_connection(ld);
	  }
    }

    if (!*ce_ap) {
        if (!sav_errno) {
            SD_setException(exception, SDStatus_FAILURE,
                    "No GlueCESEBindCEAccesspoint found.");
            sav_errno = EINVAL;
        }
        errno = sav_errno;        
        return (-1);
    }
    return (0);
}

/* generates part of bdii filter with accesscontrolbaserule */
static char *generate_acbr(const char *glueobject) {


    char tmp[64 + 511];
    char *filter = NULL, *vo = NULL;
    int nb_fqan, filterlen, i;

    if ((vo = vo_global) == NULL)
        return (NULL);

    nb_fqan = fqan_size_global;
    filterlen = nb_fqan * (sizeof (tmp)) + 3 * (64 + 255);

    if ((filter = (char *) calloc(filterlen, sizeof (char))) == NULL)
        return (NULL);

    snprintf(filter, filterlen,
            "(| (%sAccessControlBaseRule=VO:%s) (%sAccessControlBaseRule=%s) (%sAccessControlRule=%s)",
            glueobject, vo, glueobject, vo, glueobject, vo);

    for (i = 0; i < nb_fqan; ++i) {
     if(fqan_global && fqan_global[i]){
        snprintf(tmp, sizeof (tmp), " (%sAccessControlBaseRule=VOMS:%s)", glueobject, fqan_global[i]);
        strcat(filter, tmp);
	}
    }

    strcat(filter, ")");
    return (filter);
}

/**
 *  get from the BDII the endpoint for the LFC
 */
int sd_get_lfc_endpoint(char **lfc_endpoint) {

    char *filter_tmp = NULL, *query;
    SDException *exception = NULL;
    LDAPMessage *reply=NULL;
    LDAPMessage *entry=NULL;
    LDAP *ld;
    int rc = 0;
    int sav_errno = 0;
    struct berval **value;

    *lfc_endpoint = NULL;

    if ((filter_tmp = generate_acbr("GlueService")) == NULL)
        return (-1);

    query = g_strdup_printf(QUERY_NAME, QUERY_LFC_ENDPOINT, filter_tmp);
    if (!filter_tmp){
        free(filter_tmp);
    }
    else{
        g_free(query);
        return -1;
    }

    if (!exception)
        exception = g_newa(SDException, 1);


    while (!*lfc_endpoint && !sav_errno) {
    
        /* returns an LDAP reply */
        if ((ld = get_connection(exception)) == NULL) {
	   g_free(query);
           sav_errno = errno;
           return -1;
        }

    
        rc = connect_search_ldap(query, lfc_endpoint_attrs, ld, exception, &reply);
        if (rc != LDAP_SUCCESS) {
	    if(rc > 0){
                ldap_msgfree(reply);
	        close_connection(ld);
	    }
            sav_errno = errno;
            break;
        }

        if ((entry = ldap_first_entry(ld, reply)) && (value = ldap_get_values_len(ld, entry, ATTR_ENDPOINT))) {
            if (value && value[0]->bv_val) {
                if ((*lfc_endpoint = strdup(value[0]->bv_val)) == NULL)
                    sav_errno = errno ? errno : ENOMEM;
                ldap_value_free_len(value);
            }
        } else {
            SD_setException(exception, SDStatus_FAILURE, "No LFC Endpoint found");
        }
        if(rc > 0){
           ldap_msgfree(reply);
           close_connection(ld);
	}
    }

    if (!*lfc_endpoint) {
        if (!sav_errno) {

            SD_setException(exception, SDStatus_FAILURE, "No LFC Endpoint found");
            sav_errno = EINVAL;
        }
        errno = sav_errno;
        return (-1);
    }

    return 0;
}

static int get_voinfo(const char *host, const char *spacetokendesc, char **sa_path, char **salocalid) {

    static char sa_path_atnm[] = "GlueVOInfoPath";
    static char sa_key_atnm[] = "GlueChunkKey";
    static char *template = "(&%s(GlueVOInfoTag=%s)(GlueChunkKey=GlueSEUniqueID=%s))";
    static char *template2 = "(&%s(!(GlueVOInfoTag=*))(GlueChunkKey=GlueSEUniqueID=%s))";
    static char *attrs[] = {sa_key_atnm, sa_path_atnm, NULL};
    int i = 0;
    LDAPMessage *entry = NULL;
    char *filter=NULL, *filter_tmp=NULL;
    LDAP *ld;
    int rc = 0;
    LDAPMessage *reply=NULL;
    struct berval **value=NULL;
    int sav_errno = 0;
    SDException *exception = NULL;

    if (!exception)
        exception = g_newa(SDException, 1);


    if (!host || !sa_path || !salocalid) {
        SD_setException(exception, SDStatus_FAILURE, "Invalid Arguments.");
        errno = EINVAL;
        return (-1);
    }

    if (strlen(host) > 255) {
        SD_setException(exception, SDStatus_FAILURE, "Hostname too long");
        errno = ENAMETOOLONG;
        return (-1);
    }

    if ((filter_tmp = generate_acbr("GlueVOInfo")) == NULL)
        return (-1);

    if (spacetokendesc)
        rc = asprintf(&filter, template, filter_tmp, spacetokendesc, host);
    else
        rc = asprintf(&filter, template2, filter_tmp, host);

   if(filter_tmp)
    free(filter_tmp);
    
    if (rc < 0) return (-1);
    	*sa_path = *salocalid = NULL;

    while (!*sa_path && !*salocalid && !sav_errno) {

       /* returns an LDAP connection */
       if ((ld = get_connection(exception)) == NULL) {
          if(filter)
           	g_free(filter);
          sav_errno = errno;
          return -1;
       }
       
        rc = connect_search_ldap(filter, attrs, ld, exception, &reply);
        if (rc != LDAP_SUCCESS) {
	  if(rc > 0){
	    ldap_msgfree(reply);
	    close_connection(ld);
	    }
          sav_errno = errno;
          break;
        }

        if ((entry = ldap_first_entry(ld, reply))) {
            if ((value = ldap_get_values_len(ld, entry, sa_path_atnm)) != NULL) {
                if (value && value[i] && value[i]->bv_val) {
                    if ((*sa_path = strdup(value[i]->bv_val)) == NULL)
                        sav_errno = errno ? errno : ENOMEM;
		    if(value)	
                        ldap_value_free_len(value);
                }
            } else if ((value = ldap_get_values_len(ld, entry, sa_key_atnm)) != NULL) {
                if (value && value[i] && value[i]->bv_val) {
                    for (i = 0; value[i]->bv_val && !*salocalid && !rc; ++i) {
                        if (strncmp(value[i]->bv_val, "GlueSALocalID=", 14) == 0)
                            if ((*salocalid = strdup(value[i]->bv_val + 14)) == NULL)
                                sav_errno = errno ? errno : ENOMEM;
                    }
		    if(value)	
                       ldap_value_free_len(value);
                }
            } else {
                if (spacetokendesc)
                    SD_setException(exception, SDStatus_FAILURE, "GlueVOInfo for tag  and SE wrongly published");
                else
                    SD_setException(exception, SDStatus_FAILURE, "GlueVOInfo for SE (with no tag) wrongly published");
            }

        } else {
            if (spacetokendesc)
                SD_setException(exception, SDStatus_FAILURE, "[INFO] no GlueVOInfo information found about tag  and SE");
            else
                SD_setException(exception, SDStatus_FAILURE, "[INFO] no GlueVOInfo information found about SE  (with no tag)");
        }
	if(rc > 0){
           ldap_msgfree(reply);
           close_connection(ld);        
	}
    }


    if (!*sa_path && !*salocalid) {
        if (!sav_errno) {
            if (spacetokendesc) {
                SD_setException(exception, SDStatus_FAILURE, "[INFO] no GlueVOInfo information found about tag and SE ");
            } else {
                SD_setException(exception, SDStatus_FAILURE, "[INFO] no GlueVOInfo information found about SE  (with no tag) ");
                sav_errno = EINVAL;
            }
        }
        errno = sav_errno;
        return (-1);
    }

    return (0);
}

/**
 *  Get from the BDII the SAPath
 *  */

static int sd_get_storage_path_local(const char *host, const char *salocalid, char **sa_path, char **sa_root) {

    LDAPMessage *entry=NULL, *reply=NULL;
    char *query=NULL;
    struct berval **value;
    LDAP *ld;
    int sav_errno = 0;
    SDException *exception = NULL;
    int ret = 0;
    char *vo = NULL;

    if (!exception)
        exception = g_newa(SDException, 1);

    if ((vo = vo_global) == NULL)
        return (-1);

    if (salocalid == NULL) {
        salocalid = vo;
    }


    if (!host || !sa_path || !sa_root) {
        errno = EINVAL;
        SD_setException(exception, SDStatus_FAILURE, "Invalid Arguments.");
        return (-1);
    }

    *sa_path = *sa_root = NULL;

    query = g_strdup_printf(QUERY_SA_PATH, salocalid, host);

    while (!*sa_path && !*sa_root && !sav_errno) {

       if ((ld = get_connection(exception)) == NULL) {
          if(query)
	    g_free(query);
          sav_errno = errno;
          return -1;
        }

        ret = connect_search_ldap(query, sa_path_attrs, ld, exception, &reply);
        if (ret != LDAP_SUCCESS) {
	   if(ret > 0){
  	      ldap_msgfree(reply);
  	      close_connection(ld);
	   }
           sav_errno = errno;
           break;
        }

        if ((entry = ldap_first_entry(ld, reply))) {
            value = ldap_get_values_len(ld, entry, ATTR_SA_PATH_ATNM);
            if (value && value[0]->bv_val) {
                if ((strncmp(value[0]->bv_val, vo, strlen(vo)) == 0)
                        && (*(value[0]->bv_val + strlen(vo)) == ':')) {
                    if ((*sa_path = g_strdup(value[0]->bv_val + strlen(vo) + 1)) == NULL)
                        sav_errno = errno ? errno : ENOMEM;
		    if(value)	
                        ldap_value_free_len(value);
                }

            } else if ((value = ldap_get_values_len(ld, entry, ATTR_SA_ROOT_ATNM))
                    != NULL) {
	     if (value && value[0]->bv_val) {
                if ((*sa_root = strdup(value[0]->bv_val + strlen(vo) + 1))
                        == NULL) {
                    sav_errno = errno ? errno : ENOMEM;
                    if(value)	
                        ldap_value_free_len(value);
                }
	     }
            } else {
                SD_setException(exception, SDStatus_FAILURE,
                        "Both SAPath and SARoot are not set.");
            }
        } else {
            SD_setException(exception, SDStatus_FAILURE,
                    "No GlueSA information found about VO and SE.");

        }
	if(ret > 0){
	    ldap_msgfree(reply);
           close_connection(ld);
	}
    }

    if (!*sa_path && !*sa_root) {
        if (!sav_errno) {
            SD_setException(exception, SDStatus_FAILURE,
                    "No GlueSA information found about VO and SE.");
            sav_errno = EINVAL;

        }
        errno = sav_errno;	
        return (-1);
    }
    return (0);
}

int sd_get_storage_path(const char *host, const char *spacetokendesc,
        char **sa_path, char **sa_root) {

    char *salocalid = NULL;
    int rc = 0;
    SDException *exception = NULL;


    if (!exception)
        exception = g_newa(SDException, 1);

    if (!host || !sa_path || !sa_root) {

        SD_setException(exception, SDStatus_FAILURE, "Invalid Arguments.");
        errno = EINVAL;
        return (-1);
    }


    *sa_path = *sa_root = NULL;

    rc = get_voinfo(host, spacetokendesc, sa_path, &salocalid);



    if (!*sa_path) {
        rc = sd_get_storage_path_local(host, salocalid, sa_path, sa_root);
    }


    return (rc);
}

/**
 *  get from the BDII the SE type ( disk, srm_v1 )
 *
 *  */

int sd_get_se_types_and_endpoints(const char *host, char ***se_types,
        char ***se_endpoints) {
    char host_tmp[255];
    int len_tmp;
    char *port;
    LDAPMessage *entry = NULL;
    
    LDAP *ld;
    SDException *exception = NULL;

    char *query = NULL;
    int i, nbentries = 0, n = 0, rc = 0;
    char **sep = NULL, **stp = NULL, **st = NULL, **sv = NULL, **ep = NULL;
    struct berval **value = NULL;

    int sav_errno = 0;

    *se_types = NULL;
    *se_endpoints = NULL;
    len_tmp = strlen(host);

    strncpy(host_tmp, host, 255);


    if (!exception)
        exception = g_newa(SDException, 1);

    if ((port = strchr(host_tmp, ':')) != NULL)
        *port = 0;
 
       LDAPMessage *reply = NULL;
 
       query = g_strdup_printf(QUERY_TYPE_ENDPOINTS, host_tmp, host_tmp, port == NULL ? "" : port + 1);

    while (nbentries < 1 && !sav_errno) {

       /* returns an LDAP connection*/
       if ((ld = get_connection(exception)) == NULL) {
         if(query)
         	g_free(query);
         sav_errno = errno;
         return -1;
       }

        rc = connect_search_ldap(query, NULL, ld, exception, &reply);
        if (rc != LDAP_SUCCESS) {
	    if(rc > 0){
	    	ldap_msgfree(reply);
  	        close_connection(ld);
	    }
            sav_errno = errno;
            rc = -1;
            break;
        }

        if ((nbentries = ldap_count_entries(ld, reply)) < 1) {
            SD_setException(exception, SDStatus_FAILURE, "No entries for host.");          
            if(rc > 0){
	    	ldap_msgfree(reply);	    
                close_connection(ld);
	    }
            break;
        }

        ++nbentries;
        if ((st = calloc(nbentries, sizeof (char *))) == NULL || (sv = calloc(
                nbentries, sizeof (char *))) == NULL || (ep = calloc(nbentries,
                sizeof (char *))) == NULL || (stp = calloc(nbentries,
                sizeof (char *))) == NULL || (sep = calloc(nbentries,
                sizeof (char *))) == NULL) {
            if(rc > 0){
	    	ldap_msgfree(reply);
                close_connection(ld);
		}
            sav_errno = errno ? errno : ENOMEM;
            rc = -1;
            break;
        }
        for (entry = ldap_first_entry(ld, reply); entry != NULL && rc == 0; entry
                = ldap_next_entry(ld, entry)) {

            if ((value = ldap_get_values_len(ld, entry, ATTR_ENDPOINT)) != NULL) {
                if (value && value[0]->bv_val) {
                    // GlueService entry
                    ep[n] = strdup(value[0]->bv_val);
		    if(value)
                    	ber_bvecfree(value);

                    if ((value = ldap_get_values_len(ld, entry, ATTR_TYPE)) == NULL) {
                        if(value && ep[n]){
				ber_bvecfree(value);
                        	free(ep[n]);
                        	continue;
			}
                    }
		    if(st)
                    	st[n] = strdup(value[0]->bv_val);
		    if(value)
                    	ber_bvecfree(value);

                    if ((value = ldap_get_values_len(ld, entry, ATTR_VERSION))
                            == NULL) {
                        if(value && ep[n] && st[n]){
			ber_bvecfree(value);
                        free(ep[n]);
                        free(st[n]);
                        continue;
			}
                    }
		    if(sv && value && value[0]->bv_val)
                    	sv[n] = strdup(value[0]->bv_val);
		    if(value)
                    	ber_bvecfree(value);
                    ++n;
                }
            } else {
                if (port == NULL) {
                    // If port is not yet defined in host_tmp, and is available
                    // it is copied to host_tmp buffer
                    // ... But it will only be used if there is no GlueService entry
		    if(value)
                    	ber_bvecfree(value);

                    value = ldap_get_values_len(ld, entry, ATTR_SE_TYPE_ATPT);
                    if (value == NULL) {
                        if (nbentries == 1) {
                            sav_errno = EINVAL;
                            SD_setException(exception, SDStatus_FAILURE,
                                    "No port published in the BDII");
                        }
                    } else if ( value && value[0]->bv_val && (len_tmp + strlen(value[0]->bv_val) + 1 < 255)) {
                        port = host_tmp + len_tmp;
                        strcpy(port + 1, value[0]->bv_val);
                    } else {
		        if(value)
                        	ber_bvecfree(value);
                        sav_errno = ENAMETOOLONG;
                        break;
                    }

                }
		if(value)
                	ber_bvecfree(value);
            }
        }
	if(rc > 0){
        	ldap_msgfree(reply);
                close_connection(ld);
	}
    }




    if (rc < 0) {
        for (i = 0; i < n; ++i) {
            if (st && st[i])
                free(st[i]);
            if (sv && sv[i])
                free(sv[i]);
            if (ep && ep[i])
                free(ep[i]);
        }
        if (st)
            free(st);
        if (sv)
            free(sv);
        if (ep)
            free(ep);
        if (stp)
            free(stp);
        if (sep)
            free(sep);
        errno = sav_errno ? sav_errno : EINVAL;
    } else {
        if (n > 0) {
            // If there are GlueServices entries...
            for (i = 0; i < n; ++i) {
                if ((strcmp(st[i], "srm_v1") == 0 || strcmp(st[i], "srm_v2")
                        == 0) && ep[i]) {
                    stp[i] = strdup(st[i]);
                    sep[i] = strdup(ep[i]);
                } else if ((strcasecmp(st[i], "SRM") == 0) && (strncmp(sv[i],
                        "1.", 2)) == 0 && ep[i]) {
                    stp[i] = strdup("srm_v1");
                    sep[i] = strdup(ep[i]);
                } else if ((strcasecmp(st[i], "SRM") == 0) && (strncmp(sv[i],
                        "2.2", 3)) == 0 && ep[i]) {
                    stp[i] = strdup("srm_v2");
                    sep[i] = strdup(ep[i]);
                }
                free(st[i]);
                free(sv[i]);
                free(ep[i]);
            }
            free(st);
            free(sv);
            free(ep);
        } else if (n == 0 && port != NULL && stp && sep) {
            // There were no GlueService entry...
            // ... so endpoint is hostname:port, and type is disk
            *port = ':';
            stp[0] = strdup("disk");
            sep[0] = strdup(host_tmp);
        } else {
            free(stp);
            free(sep);
            rc = -1;
	    fprintf(stderr, "No available information\n");
            errno = EINVAL;

        }

        if (rc == 0) {
            *se_types = stp;
            *se_endpoints = sep;
        }
    }

    return (rc);
}

/* get from the BDII the list of supported protocols with their associated
 * port number
 */
int sd_get_seap_info(const char *host, char ***access_protocol, int **port) {
    char **ap = NULL, *attr=NULL, *query=NULL;
    struct berval **value=NULL;
    BerElement *ber = NULL;
    int nbentries, i, n, rc = 0;
    int *pn = NULL;
    LDAPMessage *entry, *reply = NULL;
    LDAP *ld = NULL;
    SDException *exception = NULL;
    int sav_errno = 0;

    /*prepare the query */
    query = g_strdup_printf(QUERY_PROTO_PORT_TYPE, host);

    if (!exception)
        exception = g_newa(SDException, 1);

    while (n < 1 && !sav_errno) {

        /* returns an LDAP connection */
        if ((ld = get_connection(exception)) == NULL) {
		if(query)
			g_free(query);
   		sav_errno = errno;
         	return -1;
        }

        /*send query to bdii server*/
        rc = connect_search_ldap(query, seap_attrs, ld, exception, &reply);
        if (rc != LDAP_SUCCESS) {
            if(rc > 0){
	       ldap_msgfree(reply);
	       close_connection(ld);
	    }
            sav_errno = errno;
            break;
        }
        /*count entry # */
        nbentries = ldap_count_entries(ld, reply);
        nbentries++;
        /*allocate memory */
        if ((ap = (char **) calloc(nbentries, sizeof (char *))) == NULL || (pn
                = (int *) calloc(nbentries, sizeof (int))) == NULL) {
	    if(rc > 0){
	      ldap_msgfree(reply);
              close_connection(ld);
	    }
            sav_errno = errno ? errno : ENOMEM;
            rc = -1;
            break;
        }

        // search for the specific values
        for (entry = ldap_first_entry(ld, reply), n = 0; entry != NULL && rc
                == 0; entry = ldap_next_entry(ld, entry), ++n) {

            for (attr = ldap_first_attribute(ld, entry, &ber); attr != NULL
                    && rc == 0; attr = ldap_next_attribute(ld, entry, ber)) {

                if (!ap[n] && (value = ldap_get_values_len(ld, entry, attr))) {
                    if (value && value[0]->bv_val) {
                        /*do the comparison*/
                        if (strcmp(attr, "GlueSEAccessProtocolType") == 0) {
                            /*save value to ap*/
                            if ((ap[n] = strdup(value[0]->bv_val)) == NULL) {

                                rc = -1;
                                sav_errno = errno ? errno : ENOMEM;
                            }
                        } else {

                            pn[n] = atoi(value[0]->bv_val);
                        }
                       if(value)
                        ber_bvecfree(value);
                    }
                }
               if(attr)
                ldap_memfree(attr);
            }
           if(ber){
            ber_free(ber, 0);
            ber = NULL;
          }
        }
        if(rc > 0){
	   ldap_msgfree(reply);
           close_connection(ld);
	}
    }

    /* if something went wrong, give the memory free */
    if (rc < 0) {

        for (i = 0; i < n; ++i) {
            if (ap[i]) {
                free(ap[i]);
            }
        }
        if(ap && pn){
          free(ap);
          free(pn);
        }
        errno = sav_errno ? sav_errno : EINVAL;
    } else /*return the values via pointer*/ {
        *access_protocol = ap;
        *port = pn;
    }
    return (rc);
}

/*RLS is not used anymore by anybody*/
int sd_get_rls_endpoints(char **lrc_endpoint, char **rmc_endpoint, char *vo) {
    return (-1);
}

void SD_freeException(SDException *exception) {
    if (exception) {
        if (exception->reason &&
                exception->reason != SD_EMPTY_MESSAGE &&
                exception->reason != SD_NOMEM_MESSAGE) {
            free(exception->reason);
            exception->reason = NULL;
        }
        /* don't free exception itself, it's not on the heap */
    }
}

void SD_freeVOList(SDVOList *vos) {
    int i;

    if (vos) {
        if (vos->names) {
            for (i = 0; i < vos->numNames; ++i) {
                if (vos->names[i]) {
                    free(vos->names[i]);
                }
            }
            free(vos->names);
        }
        free(vos);
    }
}

void SD_freeServiceDetails(SDServiceDetails *serviceDetails) {
    if (serviceDetails) {
        if (serviceDetails->name) free(serviceDetails->name);
        if (serviceDetails->type) free(serviceDetails->type);
        if (serviceDetails->endpoint) free(serviceDetails->endpoint);
        if (serviceDetails->version) free(serviceDetails->version);
        if (serviceDetails->site) free(serviceDetails->site);
        if (serviceDetails->wsdl) free(serviceDetails->wsdl);
        if (serviceDetails->administration) free(serviceDetails->administration);
        if (serviceDetails->vos) SD_freeVOList(serviceDetails->vos);
        if (serviceDetails->associatedServices) SD_freeServiceList(serviceDetails->associatedServices);
        if (serviceDetails->data) SD_freeServiceDataList(serviceDetails->data);
        free(serviceDetails);
    }
}

/**
 * Default implementation of the freeServiceDataList method.
 *
 * @param  A SD_freeServiceDataList.
 */

void SD_freeServiceDataList(SDServiceDataList *serviceDataList) {
    int i;

    if (serviceDataList) {
        if (serviceDataList->items) {
            for (i = 0; i < serviceDataList->numItems; ++i) {
                if (serviceDataList->items[i].key) {
                    free(serviceDataList->items[i].key);
                }
                if (serviceDataList->items[i].value) {
                    free(serviceDataList->items[i].value);
                }
            }
            free(serviceDataList->items);
        }
        free(serviceDataList);
    }
}

/**
 * Default implementation of the freeService method.
 *
 * @param  A SD_freeService.
 */

void SD_freeService(SDService *service) {
    if (service) {
        if (service->name)
            free(service->name);
        if (service->type)
            free(service->type);
        if (service->endpoint)
            free(service->endpoint);
        if (service->version)
            free(service->version);
        free(service);
    }
}

/**
 * Default implementation of the freeServiceList method.
 *
 * @param  As SD_freeServiceList.
 */

void SD_freeServiceList(SDServiceList *serviceList) {
    int i;

    if (serviceList) {
        if (serviceList->services) {
            for (i = 0; i < serviceList->numServices; ++i) {
                if (serviceList->services[i]) {
                    SD_freeService(serviceList->services[i]);
                }
            }
            free(serviceList->services);
        }
        free(serviceList);
    }
}

/*
This is required by Savannah task  31706: lcg-utils/GFAL should check GlueSEStatus.
If it is "Production" or missing then the API function returns true, else false. 
The parameter passed is the FQN of the SE
*/

int sd_is_online(char *host) {
    char host_tmp[255];
    LDAPMessage *reply = NULL;
    LDAP *ld;
    SDException *exception = NULL;

    char *query;
    int rc = 0;
    int is_production;

    strncpy(host_tmp, host, 255);

    if (!exception)
        exception = g_newa(SDException, 1);

    query = g_strdup_printf(QUERY_SE_STATUS, host_tmp);

    if ((ld = get_connection(exception)) == NULL) {
    	g_free(query);
        return -1;
    }

    rc = connect_search_ldap(query, NULL, ld, exception, &reply);

    if (rc != LDAP_SUCCESS) {
        if(rc > 0)  {
		ldap_msgfree(reply);
        	close_connection(ld);
	}
        SD_setException(exception, SDStatus_FAILURE, "No available information");
        return -1;
    }


    if ((ldap_count_entries(ld, reply)) < 1)
        is_production = 0;
    else
        is_production = 1;

    if(rc > 0){
    	ldap_msgfree(reply);
    	close_connection(ld);
    }

    if (is_production == 0)
        return (-1);

    return (0);
}
